/*
** Copyright (C) 1999 Eleftherios Gkioulekas <lf@amath.washington.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#ifndef __defined_mu_h
#define __defined_mu_h

/* I am not putting an effort to be portable to lame systems that 
   don't support stdarg.h 
   If the need every arises to support such a system, then it is not
   very difficult to add support for it */
#include <stdio.h>
#include <stdarg.h>
#include <setjmp.h>
#include <assert.h>

/*
 * -------------------------
 * Mathutil error handling 
 * -------------------------
 */

/* The following global variable defines how mathutil should
   react to errors */
extern int mu_error_status;
extern int mu_warnings_off;

/* Go to surfer mode with the error messages */
void mu_error_surfer_mode (void);
void mu_error_set_mode (char *message_str, char *warning_str, char *error_str);

/* Portable strerror for C library errors */
char * mu_libc_strerror (int errnum);

/* Routines for handling non-local exits */
void mu_error_push_jump (jmp_buf jmp);
void mu_error_pop_jump (void);
void mu_error_continue_next_jump (void);

/* Routines for setting the error reporter */
typedef void mu_error_reporter_t (char *prefix, int errnum,
				  char *filename, int line, 
				  char *fmt, va_list args);
mu_error_reporter_t * 
mu_set_error_reporter (mu_error_reporter_t * handler);

/* These routines set the error stream for the default error reporter */
FILE * mu_set_error_stream (FILE *stream);
FILE * mu_set_error_stream_by_filename (char *filename);

/* Routines for throwing Mathutil errors, warnings, or messages */
void mu_error_stream_write (char *prefix, int errnum, 
			    char *filename, int line,
			    char *fmt, va_list args);
void mu_error (int status, int errnum, char *fmt, ...);
void mu_error_at_line (int status, int errnum, char *filename, int line, 
		       char *fmt, ...);
void mu_warning (int status, int errnum, char *fmt, ...);
void mu_message (char *fmt, ...);

/* Convert Mathutil error code to string */
char * mu_strerror (int errnum);

/* 
 * --------------------
 * Mathutil error list 
 * --------------------
 */

/* This is the same error list used by the GNU Scientific Library */
enum { 
  MU_EDOM     = 1,   /* input domain error, e.g sqrt(-1) */
  MU_ERANGE   = 2,   /* output range error, e.g. exp(1e100) */
  MU_EFAULT   = 3,   /* invalid pointer */
  MU_EINVAL   = 4,   /* invalid argument supplied by user */
  MU_EFAILED  = 5,   /* generic failure */
  MU_ESANITY  = 6,   /* sanity check failed - shouldn't happen */
  MU_ENOMEM   = 7,   /* malloc failed */
  MU_EIO      = 8,   /* input/output error */
  MU_EBADFUNC = 9,   /* problem with user-supplied function */
  MU_ERUNAWAY = 10,  /* iterative process is out of control */
  MU_EMAXITER = 11,  /* exceeded max number of iterations */
  MU_EZERODIV = 12,  /* tried to divide by zero */
  MU_EBADTOL  = 13,  /* user specified an invalid tolerance */
  MU_ETOL     = 14,  /* failed to reach the specified tolerance */
  MU_EUNDRFLW = 15,  /* underflow */
  MU_EOVRFLW  = 16,  /* overflow  */
  MU_ELOSS    = 17,  /* loss of accuracy */
  MU_EROUND   = 18,  /* failed because of roundoff error */
  MU_EBADLEN  = 19,  /* matrix, vector lengths are not conformant */
  MU_ESING    = 20,  /* apparent singularity detected */
  MU_EUNSUP   = 21,  /* requested feature is not supported by the hardware */
  MU_EUNIMPL  = 22,  /* requested feature not (yet) implemented */
  MU_EINDEX   = 23   /* smart array index error */
} ;

/* 
 * -------------------
 * Variadic functions
 * -------------------
 */

/* Handler for variadic arguments */
typedef int mu_var_handler_t (void *clientdata, char *spec, va_list *arg);

/* Table of handlers */
struct mu_var_handler_table;

struct mu_var_handler_table
{
  mu_var_handler_t *handler;
  void *next;
};

/* Register a handler */
void mu_var_register (struct mu_var_handler_table *table, 
		      mu_var_handler_t *handler);
void mu_var_call (struct mu_var_handler_table *table, void *clientdata,
		  char *routine, char *fmt, va_list *arg);


/*
 * --------------------------------
 * Memory allocation/deallocation
 * --------------------------------
 */

void *mu_malloc (size_t len, size_t size);
void *mu_calloc (size_t len, size_t size);
void *mu_realloc (void *p, size_t len, size_t size);
void mu_free (void *p);


/*
 * ----------------------------------
 * Mathutil binary file input/output
 * -----------------------------------
 */

/*
 * Endian issues
 */

/* The following enum determines the file type */
enum {
  mu_file_big_endian    = 1,
  mu_file_little_endian = 2 
};

/* Detect the endian-ness of the current platform */
int mu_endian_detect (void);

/* Endian to descriptive string conversion */
void mu_endian_validate (int endian, char * routine);
char * mu_endian_to_string (int endian);
int mu_endian_from_string (char *str);
int mu_endian_valid_filetype_p (char *str);

/* Endian routines for swaping */
void mu_endian_swap4 (void *c);
void mu_endian_swap8 (void *c);
void mu_endian_swap4_block (void *c, size_t len);
void mu_endian_swap8_block (void *c, size_t len);

/*
 * Simple I/O
 */

/* The following enum defines the mode with which we open the file */
enum {
  mu_file_for_reading   = 1,
  mu_file_for_writing   = 2,
  mu_file_for_appending = 3
};

/* Handler for end-of-file conditions */
struct mu_file;
typedef void mu_file_eof_handler_t (struct mu_file *file);

/* The following data type holds the state of the binary file stream */
struct mu_file
{
  char *filename;         /* The filename corresponding to this file */
  int descriptor;         /* File descriptor for reading and writing */
  int endian;             /* The endianness of the file */
  int filemode;           /* The file mode */
  int swap_endian;        /* Do we need to swap for endianesss */
  int eof;                /* End-of-file flag */
  mu_file_eof_handler_t *eof_handler;
};

/* Delete a file */
void mu_delete (char *filename);

/* FIXME: TODO */
void mu_mkdir (char *filename);
void mu_rmdir (char *filename);

/* File open and close interface */
struct mu_file *mu_file_open_raw (char *filename, int filemode);
void mu_file_set_endian (struct mu_file *file, int endian);
void mu_file_assert_endian (struct mu_file *file, char *routine);
struct mu_file *mu_file_open (char *filename, int filemode);
void mu_file_close (struct mu_file *file);

/* Low-level input/output */
int mu_file_eof_p (struct mu_file *file);
void mu_file_read (struct mu_file *file, void *ptr, size_t len);
void mu_file_write (struct mu_file *file, void *ptr, size_t len);

/* End of file */
int mu_file_eof_p (struct mu_file *file);
mu_file_eof_handler_t * 
mu_file_set_eof_handler (struct mu_file *file, mu_file_eof_handler_t *handler);
mu_file_eof_handler_t * mu_file_set_default_eof_handler (struct mu_file *file);
 
/* Cookies */
void mu_file_write_cookie (struct mu_file *file, char *cookie);
int mu_file_read_cookie (struct mu_file *file, char *cookie);

/* Medium-level output */
void mu_file_write_double (struct mu_file *file, double *a, size_t n);
void mu_file_write_float (struct mu_file *file, float *a, size_t n);
void mu_file_write_int (struct mu_file *file, int *a, size_t n);

/* Medium-level input */
void mu_file_read_double (struct mu_file *file, double *a, size_t n);
void mu_file_read_float (struct mu_file *file, float *a, size_t n);
void mu_file_read_int (struct mu_file *file, int *a, size_t n);

/*
 * High-level I/O
 */

/* This is the definition of the printf handler function */
typedef int mu_file_io_handler_t (struct mu_file *file, char *spec, 
				  va_list *arg);

/* High level input and output */
void mu_file_input (struct mu_file *file, char *fmt, ...);
void mu_file_output (struct mu_file *file, char *fmt, ...);
void mu_file_register (mu_var_handler_t *input, mu_var_handler_t *output);

/*
 * --------------------------------------------------------
 * The membuf ADT - Memory buffers with reference counting
 * --------------------------------------------------------
 */

struct mu_membuf
{
  void *data;
  size_t size;
  size_t *refcount;
};

/* Memory bufffers should be created with one of the following two methods
   before they are ever used */
void mu_membuf_initialize (struct mu_membuf *buf);
struct mu_membuf mu_membuf_allocate (size_t size);
struct mu_membuf mu_membuf_attach_to (struct mu_membuf parent);

/* Manipulate a memory buffer */
void mu_membuf_release (struct mu_membuf buf);
void mu_membuf_reallocate (struct mu_membuf *buf, size_t size);
void mu_membuf_reattach_to (struct mu_membuf *buf, struct mu_membuf parent);

/* Quering a memory buffer */
int mu_membuf_references (struct mu_membuf buf);
void *mu_membuf_address (struct mu_membuf buf);

/* The null memory buffer */
extern struct mu_membuf mu_membuf_null;

/*
 * ------------------------------------
 * Declare the arrays for various types
 * ------------------------------------
 */

/* The slicing type is independent of the particular array type, so it
   needs to be defined here */
typedef enum 
{
  X_slice, Y_slice, Z_slice, 
  XY_slice, XZ_slice, YX_slice, YZ_slice, ZX_slice, ZY_slice
} mu_slice;

/* The file mu_decls.h is reused many times to declare the appropriate
   declarations for each type. The file mu_impl.ic is used as a template
   for creating the corresponding files that implement the functionaility
   being declared here. */

#define TYPE double
#include <mu_def_double.h>
#include <mu_decls.h>
#include <mu_undefs.h>
#undef TYPE

#define TYPE float
#include <mu_def_float.h>
#include <mu_decls.h>
#include <mu_undefs.h>
#undef TYPE

#define TYPE int
#include <mu_def_int.h>
#include <mu_decls.h>
#include <mu_undefs.h>
#undef TYPE

/* 
 * Typedef declarations to save typing 
 */

/* Doubles */
typedef struct mu_double_array double_array_t;
typedef struct mu_double_matrix double_matrix_t;
typedef struct mu_double_cube double_cube_t;

/* Floats */
typedef struct mu_float_array float_array_t;
typedef struct mu_float_matrix float_matrix_t;
typedef struct mu_float_cube float_cube_t;

/* Ints */
typedef struct mu_int_array int_array_t;
typedef struct mu_int_matrix int_matrix_t;
typedef struct mu_int_cube int_cube_t;


/*
 * ------------------------------------------------
 * Macros for accessing arrays, matrices and cubes
 * -------------------------------------------------
 */

/* The following functions are used to handle index errors */
char *mu_error_index_1 (char *filename, int line, char *a, int i);
char *mu_error_index_2 (char *filename, int line, char *a, 
			int i, int j);
char *mu_error_index_3 (char *filename, int line, char *a,
			int i, int j, int k);

/* Where to write index errors */
extern FILE *mu_index_error_stream;


/* The following macros return the memory address corresponding to 
   specific indices. They are used to construct the macros that are
   actually exported to the user */
#define Aa_(a,i) ((a).cache + (a).offset + (i)*(a).iX)
#define Ma_(a,i,j) ((a).cache + (a).offset + (i)*(a).iX + (j)*(a).iY)
#define Ca_(a,i,j,k) \
((a).cache + (a).offset + (i)*(a).iX + (j)*(a).iY + (k)*(a).iZ)

/* Here are the macros that are actually exported to users */
#if MU_CHECK_RANGE
#define mu_between(i,max) ((0 <= (i)) && ((i) < (max)))
#define A_(a,i) \
(*((mu_between((i),(a).X) || mu_error_index_1 (__FILE__, __LINE__, #a, i)) ? \
   (Aa_((a),(i))) : (Aa_((a),(i)))))
#define M_(a,i,j) \
(*(((mu_between((i),(a).X) && mu_between((j),(a).Y)) || \
    mu_error_index_2 (__FILE__, __LINE__, #a,i, j)) ? (Ma_((a),(i),(j))) : \
    (Ma_((a),(i),(j)))))
#define C_(a,i,j,k) \
(*(((mu_between((i),(a).X) && mu_between((j),(a).Y) &&  \
     (mu_between((k),(a).Z))) || \
    mu_error_index_3 (__FILE__, __LINE__, #a, i, j, k)) ? \
   (Ca_((a),(i),(j),(k))) : (Ca_((a),(i),(j),(k)))))
#else
#define A_(a,i) (*(Aa_((a),(i))))
#define M_(a,i,j) (*(Ma_((a),(i),(j))))
#define C_(a,i,j,k) (*(Ca_((a),(i),(j),(k))))
#endif /* MU_CHECK_RANGE */

/* The following macros can be used to access arrays from offset 1 */
#define A1_(a,i) A_(a,(i-1))
#define M1_(a,i,j) M_(a,(i-1),(j-1))
#define C1_(a,i,j,k) C_(a,(i-1),(j-1),(k-1))

/* The following macros provide shortcuts to numerical loops */
#define loop(COUNTER,BEGIN,END) \
 for (COUNTER = BEGIN; COUNTER <= END; COUNTER += 1)

#define loop_with_step(COUNTER,BEGIN,END,STEP) \
 for (COUNTER = BEGIN; COUNTER <= END; COUNTER += STEP)

#define loop_array(COUNTER,ARRAY) \
 for (COUNTER = 0; COUNTER <= ((ARRAY).X - 1); COUNTER += 1 )

#define loop_matrix(I,J,MATRIX) \
 loop(J,0,(MATRIX).Y-1) loop(I,0,(MATRIX).X-1)

#define loop_cube(I,J,K,CUBE) \
 loop(K,0,(CUBE).Z-1) loop(J,0,(CUBE).Y-1) loop(I,0,(CUBE).X-1)

#endif /* __defined_mu_h */
