/*
** Copyright (C) 1999 Eleftherios Gkioulekas <lf@amath.washington.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

/*
 * Array, matrix and cube generic declarations. 
 * This file is included multiple times but with different definition
 * of TYPE
 */

/*
 * ---------
 * 1d arrays
 * ---------
 */

struct mu_array
{
  struct mu_membuf buf;
  TYPE *cache;
  size_t offset,X,iX;
};

/* Constructors and destructors */
struct mu_array mu_array_initialize (void);
struct mu_array mu_array_alloc (size_t x);
struct mu_array mu_array_alloc_from_buffer (struct mu_membuf buf, 
					    size_t offset, 
					    size_t X, size_t iX);
struct mu_array mu_array_alias (struct mu_array a);
void mu_array_free (struct mu_array a);

/* Reallocators */
void mu_array_realloc (struct mu_array *a, size_t X);
void mu_array_realloc_from_buffer (struct mu_array *a, struct mu_membuf buf,
				   size_t offset, size_t X, size_t iX);
void mu_array_realias (struct mu_array *a, struct mu_array b);

/* Make array contiguous */
int mu_array_contiguous_p (struct mu_array a);

/* get/check */
TYPE * mu_array_address (struct mu_array a);
TYPE mu_array_get (struct mu_array a, size_t i);
void mu_array_set (struct mu_array a, size_t i, TYPE val);

/* 
 * ----------
 * 2d arrays
 * ----------
 */

struct mu_matrix
{
  struct mu_membuf buf;
  TYPE *cache;
  size_t offset,X,iX,Y,iY;
};

/* Constructors and destructors */
struct mu_matrix mu_matrix_initialize (void);
struct mu_matrix mu_matrix_alloc (size_t x, size_t y);
struct mu_matrix mu_matrix_alloc_from_buffer (struct mu_membuf buf,
					      size_t offset,
					      size_t X, size_t iX,
					      size_t Y, size_t iY);
struct mu_matrix mu_matrix_alias (struct mu_matrix a);
void mu_matrix_free (struct mu_matrix a);

/* Reallocators */
void mu_matrix_realloc (struct mu_matrix *a, size_t X, size_t Y);
void mu_matrix_realloc_from_buffer (struct mu_matrix *a, struct mu_membuf buf,
				    size_t offset, size_t X, size_t iX,
				    size_t Y, size_t iY);
void mu_matrix_realias (struct mu_matrix *a, struct mu_matrix b);

/* Make matrix contiguous */
int mu_matrix_contiguous_p (struct mu_matrix a);

/* Get and set */
TYPE *mu_matrix_address (struct mu_matrix a);
TYPE mu_matrix_get (struct mu_matrix a, size_t i, size_t j);
void mu_matrix_set (struct mu_matrix a, size_t i, size_t j, TYPE val);

/* Transpose matrices */
int mu_matrix_transpose_contiguous_p (struct mu_matrix a);
void mu_matrix_transpose (struct mu_matrix *a);

/*
 * ----------
 * 3d arrays
 * ----------
 */

struct mu_cube
{
  struct mu_membuf buf;
  TYPE *cache;
  size_t offset,X,iX,Y,iY,Z,iZ;
};

/* Constructors and destructors */
struct mu_cube mu_cube_initialize (void);
struct mu_cube mu_cube_alloc (size_t x, size_t y, size_t z);
struct mu_cube mu_cube_alloc_from_buffer (struct mu_membuf buf,
					  size_t offset,
					  size_t X, size_t iX,
					  size_t Y, size_t iY,
					  size_t Z, size_t iZ);
struct mu_cube mu_cube_alias (struct mu_cube a);
void mu_cube_free (struct mu_cube a);

/* Reallocators */
void mu_cube_realloc (struct mu_cube *a, size_t X, size_t Y, size_t Z);
void mu_cube_realloc_from_buffer (struct mu_cube *a, struct mu_membuf buf,
				  size_t offset, size_t X, size_t iX,
				  size_t Y, size_t iY, size_t Z, size_t iZ);
void mu_cube_realias (struct mu_cube *a, struct mu_cube b);

/* Check if cube is contiguous */
int mu_cube_contiguous_p (struct mu_cube a);

/* Get and set */
TYPE *mu_cube_address (struct mu_cube a);
TYPE mu_cube_get (struct mu_cube a, size_t i, size_t j, size_t k);
void mu_cube_set (struct mu_cube a, size_t i, size_t j, size_t k, TYPE val);

/*
 * ---------------------
 * Helper de-allocators 
 * ---------------------
 */

void mu_array_free_many (int k, ...);
void mu_matrix_free_many (int k, ...);
void mu_cube_free_many (int k, ...);

/* 
 * --------------------------------
 * Derived arrays, matrices, cubes
 * --------------------------------
 */

/* Nulls */
extern struct mu_array mu_array_null;
extern struct mu_matrix mu_matrix_null;
extern struct mu_cube mu_cube_null;

/* Subsidiary */
void mu_array_sub (struct mu_array *a, struct mu_array b, 
		   size_t i1, size_t i2);
void mu_matrix_sub (struct mu_matrix *a, struct mu_matrix b,
		    size_t i1, size_t i2, size_t j1, size_t j2);
void mu_cube_sub (struct mu_cube *a, struct mu_cube b,
		  size_t i1, size_t i2, size_t j1, size_t j2,
		  size_t k1, size_t k2);

/* Interlaces */
void mu_array_interlaced (struct mu_array a, struct mu_array *a1,
			  struct mu_array *a2);
void mu_matrix_interlaced (struct mu_matrix a, struct mu_matrix *a1,
			   struct mu_matrix *a2);
void mu_cube_interlaced (struct mu_cube a, struct mu_cube *a1, 
			 struct mu_cube *a2);

/* 2d -> 1d slicing */
void mu_matrix_to_array (struct mu_array *b, struct mu_matrix a, 
			 mu_slice slice, int i);

/* 3d -> 1d slicing */
void mu_cube_to_array (struct mu_array *b, struct mu_cube a, 
		       mu_slice slice, int i, int j);

/* 3d -> 2d slicing */
void mu_cube_to_matrix (struct mu_matrix *b, struct mu_cube a,
			mu_slice slice, int i);

/*
 * -----------------
 * Smart array I/O 
 * -----------------
 */

/* Routines to write smart arrays to a binary file */
void mu_array_write (struct mu_file *file, struct mu_array a);
void mu_matrix_write (struct mu_file *file, struct mu_matrix a);
void mu_cube_write (struct mu_file *file, struct mu_cube a);

/* Routines to read a smart array from a binary file */
void mu_array_read (struct mu_file *file, struct mu_array *a, int load);
void mu_matrix_read (struct mu_file *file, struct mu_matrix *a, int load);
void mu_cube_read (struct mu_file *file, struct mu_cube *a, int load);

