/*
** Copyright (C) 1999 Eleftherios Gkioulekas <lf@amath.washington.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
# include <stdlib.h>
#endif
#if HAVE_MALLOC_H
# include <malloc.h>
#endif
#include <errno.h>
#include <mu.h>

/* This might become necessary later, if we add support for varagrs */
#define VA_START(args, lastargs) va_start(args, lastargs)


/*
 * -------------
 * Error modes
 * -------------
 */

/* Prefixes for warning, error and message */
char *mu_prefix_warning = "WARNING";
char *mu_prefix_error   = "ERROR";
char *mu_prefix_message = "MESSAGE";

/* Set error mode */
void mu_error_set_mode (char *message_str, char *warning_str, char *error_str)
{
  /* Automagically resizing buffers that persist */
  static char *buf1 = NULL; static int buf1_len = 0;
  static char *buf2 = NULL; static int buf2_len = 0;
  static char *buf3 = NULL; static int buf3_len = 0;

  /* Get the lengths of the prefixes */
  int message_len = strlen (message_str) + 1;
  int warning_len = strlen (warning_str) + 1;
  int error_len   = strlen (error_str)   + 1;

  if (buf1_len < message_len)
    {
      buf1_len = message_len;
      buf1 = (char *) mu_realloc (buf1, buf1_len, sizeof (char));
    }

  if (buf2_len < warning_len)
    {
      buf2_len = warning_len;
      buf2 = (char *) mu_realloc (buf2, buf2_len, sizeof (char));
    }

  if (buf3_len < error_len)
    {
      buf3_len = error_len;
      buf3 = (char *) mu_realloc (buf3, buf3_len, sizeof (char));
    }

  strcpy (buf1, message_str);
  strcpy (buf2, warning_str);
  strcpy (buf3, error_str);
  mu_prefix_message = buf1;
  mu_prefix_warning = buf2;
  mu_prefix_error   = buf3;
  
}

/* Set surfer mode */
void mu_error_surfer_mode (void)
{
  mu_error_set_mode ("Dude, ya got a message",
		     "Bad news Dude",
		     "Major wipe out Dude");
  
}

/*
 * ---------------------------------------
 * Portable strerror for C library errors 
 * ---------------------------------------
 */

#if HAVE_STRERROR
# ifndef strerror
char *strerror ();
# endif
#endif

char * mu_libc_strerror (int errnum)
{
#if HAVE_STRERROR
  return strerror (errnum);
#else
  static char lose[] = "Unknown system error";
  return lose;
#endif
}

/*
 * ---------------------------------------
 * Routines for handling non-local exits
 * ---------------------------------------
 */

/* Stack for storing jumo buffers */
struct mu_jump_stack
{
  jmp_buf buf;
  void *next;
};

struct mu_jump_stack *mu_jumps = NULL;
int mu_jump_status = 0;

/* Routine to copy one jump buffer into another.
   A word of explanation; Why Does It Work??
   According to the C specs, the setjmp function is capable of 
   modifying the contents of a jmp_buf if it is given one as an
   argument. The only way this is possible, is if jmp_buf is 
   actually implemented as an opaque pointer to the actual data 
   Credit for the idea: John Eaton. You rule dude!
*/

static void mu_jump_copy (jmp_buf from, jmp_buf to)
{
  int size = sizeof (jmp_buf);
  char *from_addr = (char *) from;
  char *to_addr   = (char *) to;
  while (size-- > 0) *to_addr++ = *from_addr++;
}

/* Push a jump buffer into the stack */
void mu_error_push_jump (jmp_buf jmp)
{
  struct mu_jump_stack *entry;
  entry = (struct mu_jump_stack *) 
    mu_malloc (1, sizeof (struct mu_jump_stack));
  entry->next = mu_jumps;
  mu_jumps = entry;
  mu_jump_copy (jmp, entry->buf);
}

/* Pop a jump buffer from the stack without jumping */
void mu_error_pop_jump (void)
{
  struct mu_jump_stack *entry = mu_jumps;
  assert (entry); /* entry should not be NULL if push/pop are balanced */
  mu_jumps = entry->next;
  free (entry);
}

/* Pop a jump buffer from the stack and do a long jump to it */
void mu_error_continue_next_jump (void)
{
  struct mu_jump_stack *entry = mu_jumps;

  /* If the entry is NULL, then determine what to do using the status */
  if (entry == NULL)
    {
      if (mu_jump_status) exit (mu_jump_status);
      return;
    }
  /* Otherwise, pop the entry and do a longjmp */
  else
    {
      jmp_buf jmp;
      mu_jump_copy (entry->buf, jmp);
      mu_jumps = entry->next;
      free (entry);
      longjmp (jmp, 1);
    }
}

/* Do the initial jump */
static void mu_error_start_jump (int status)
{
  mu_jump_status = status;
  mu_error_continue_next_jump ();
}

/*
 * ---------------------------------------
 * Routine for setting the error reporter
 * ---------------------------------------
 */

/* The error reporter is in charge of controlling how to report errors
   back to the user. The default reporter sends the error message to
   a stream, which can also be controlled. A non-default reporter could
   report the error differently. For example, it could use a graphical
   user interface instead, if the whole application is GUI */

/* Default error reporter */
mu_error_reporter_t * mu_error_reporter = mu_error_stream_write;

/* Change the error reporter */
mu_error_reporter_t * 
mu_set_error_reporter (mu_error_reporter_t * handler)
{
  mu_error_reporter_t *old_handler = mu_error_reporter;
  mu_error_reporter = handler;
  return old_handler;
}

/*
 * ---------------------------------------
 * Error stream and handler manipulation
 * ---------------------------------------
 */

/* Standard stream where errors should be written */
FILE *mu_error_stream = NULL;

/* Redirect the error stream */
FILE *mu_set_error_stream (FILE *stream)
{
  FILE *old_stream = mu_error_stream;
  mu_error_stream = stream;
  return old_stream;
}

/* Redirect the error stream to a specific file */
FILE *mu_set_error_stream_by_filename (char *filename)
{
  FILE *stream = fopen (filename, "w");
  if (stream == NULL)
    {
      mu_error (mu_error_status, MU_EIO, 
		"Can't open file %s for error log", filename);
    }
  return mu_set_error_stream (stream);
}

/* 
 * -----------------------------------------
 * Routines for throwing Mathutil errors
 * -----------------------------------------
 */

/* If set equal to 1 then mathutil will abort when an error happens.
   If set equal to 0 then mathutil will not abort */
int mu_error_status = 1;

/* If set equal to 1 then warnings are disabled */
int mu_warnings_off = 0;

/* This routine outputs an error message to the error stream.
   It is useful for writing error handlers */
void mu_error_stream_write (char *prefix, int errnum, 
			    char *filename, int line,
			    char *fmt, va_list args)
{
  /* Set default, if necessary */
  if (mu_error_stream == NULL) mu_error_stream = stderr;

  /* Print prefix, and filename stuff if applicable */
  if (filename) fprintf (mu_error_stream, "%s:%d: ", filename, line);
  if (prefix) fprintf (mu_error_stream, "%s: ", prefix);
  if (errnum) fprintf (mu_error_stream, "%s", mu_strerror (errnum));

  /* Print the error provided by the user */
  if (fmt)
    {
      /* If we have already printed the errnum error message, then 
	 get a new line for this error message */
      if ((prefix) && (errnum))
	{
	  fprintf (mu_error_stream, "\n%s: ", prefix);
	}
      /* Print the user error message */
#if HAVE_VPRINTF
      vfprintf (mu_error_stream, fmt, args);
#else
      _doprnt (fmt, args, mu_error_stream);
#endif
    }

  /* Print the error corresponding to errnum if appropriate */
  fprintf (mu_error_stream, "\n");
  fflush (mu_error_stream);
}

void mu_error (int status, int errnum, char *fmt, ...)
{
  /* Get the arguments */
  va_list args;
  va_start (args, fmt);

  /* Report the error to the user */
  if (mu_error_reporter)
    (*mu_error_reporter) (mu_prefix_error, errnum, NULL, 0, fmt, args);
  va_end (args);

  /* Do the appropriate non-local exit */
  mu_error_start_jump (status);
}

void mu_error_at_line (int status, int errnum, char *filename, int line,
		       char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);
  if (mu_error_reporter)
    (*mu_error_reporter) (mu_prefix_error, errnum, filename, line, fmt, args);
  va_end (args);
  mu_error_start_jump (status);
}

void mu_warning (int status, int errnum, char *fmt, ...)
{
  va_list args;
  VA_START (args, fmt);
  if ((mu_warnings_off != 0) && mu_error_reporter)
    {
      (*mu_error_reporter) (mu_prefix_warning, errnum, NULL, 0, fmt, args);
    }
  va_end (args);
}

void mu_message (char *fmt, ...)
{
  va_list args;
  VA_START (args, fmt);
  if (mu_error_reporter)
    (*mu_error_reporter) (mu_prefix_message, 0, NULL, 0, fmt, args);
  va_end (args);
}

/*
 * -------------------------
 * Error codes for Mathutil 
 * -------------------------
 */


char *mu_strerror (int errnum)
{
  switch (errnum)
    {
    case MU_EDOM:
      return "input domain error" ;
    case MU_ERANGE:
      return "output range error" ;
    case MU_EFAULT:
      return "invalid pointer" ;
    case MU_EINVAL:
      return "invalid argument supplied by user" ;
    case MU_EFAILED:
      return "generic failure" ;
    case MU_EIO:
      return "I/O failure";
    case MU_ESANITY:
      return "sanity check failed - shouldn't happen" ;
    case MU_ENOMEM:
      return "malloc failed" ;
    case MU_EBADFUNC:
      return "problem with user-supplied function";
    case MU_ERUNAWAY:
      return "iterative process is out of control";
    case MU_EMAXITER:
      return "exceeded max number of iterations" ;
    case MU_EZERODIV:
      return "tried to divide by zero" ;
    case MU_EBADTOL:
      return "specified tolerance is invalid or theoretically unattainable" ;
    case MU_ETOL:
      return "failed to reach the specified tolerance" ;
    case MU_EUNDRFLW:
      return "underflow" ;
    case MU_EOVRFLW:
      return "overflow" ;
    case MU_ELOSS:
      return "loss of accuracy" ;
    case MU_EROUND:
      return "roundoff error" ;
    case MU_EBADLEN:
      return "matrix/vector sizes are not conformant" ;
    case MU_ESING:
      return "singularity or extremely bad function behavior detected" ;
    case MU_EUNSUP:
      return "the required feature is not supported by this hardware platform";
    case MU_EUNIMPL:
      return "the requested feature is not (yet) implemented";
    case MU_EINDEX:
      return "smart array index error";
    default:
      return "unknown error code" ;
    }
}

/*
 * ----------------------------------------------
 * The following routines handle error exception 
 * for smart array indexing 
 * ----------------------------------------------
 */

/* The following errors are completely unrecoverable, so an 
   abortion is really necessary */
char *mu_error_index_1 (char *filename, int line, char *a, int i)
{
  mu_error_at_line (0, MU_EINDEX, filename, line, 
		    "offending array %s (%d)", a, i);
  abort ();
  return NULL;
}

char *mu_error_index_2 (char *filename, int line, char *a, int i, int j)
{
  mu_error_at_line (0, MU_EINDEX, filename, line,
		    "offending array %s (%d, %d)", a, i, j);
  abort();
  return NULL;
}

char *mu_error_index_3 (char *filename, int line, char *a, int i, int j, int k)
{
  mu_error_at_line (0, MU_EINDEX, filename, line,
		    "offending array %s (%d, %d, %d)", a, i, j, k);
  abort ();
  return NULL;
}

