/*
** Copyright (C) 1999 Eleftherios Gkioulekas <lf@amath.washington.edu>me
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
# include <stdlib.h>
#endif
#if HAVE_STRING_H
# include <string.h>
#else
# include <strings.h>
#endif
#include <errno.h>
#ifndef errno
extern int errno;
#endif
#include <mu.h>

#if HAVE_UNISTD_H
# include <unistd.h>
#endif
#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#if HAVE_FCNTL_H
# include <fcntl.h>
#endif
#if HAVE_SYS_FILE_H
# include <sys/file.h>
#endif

#define VA_START(args, lastarg) va_start(args, lastarg)
 
/* Delete a file */
void mu_delete (char *filename)
{
  if (remove (filename) < 0)
    {
      mu_error (mu_error_status, MU_EIO, "Failed to delete %s: %s",
		filename, mu_libc_strerror (errno));
    }
}


/* 
 * ---------------------------
 * Opening and closing files
 * ---------------------------
 */

/* Open file and set the endian mode manually. This is useful for reading
   files that were not created by a Mathutil program */
struct mu_file *mu_file_open_raw (char *filename, int filemode)
{
  /* Initialize the data structure storing the file */
  struct mu_file *file;
  int flags;

  /* Set the flags */
  if (filemode == mu_file_for_reading)
    {
      flags = O_RDONLY;
    }
  else if (filemode == mu_file_for_writing)
    {
      flags = O_WRONLY | O_CREAT | O_TRUNC;
    }
  else if (filemode == mu_file_for_appending)
    {
      flags = O_WRONLY | O_APPEND;
    }
  else
    {
      mu_error (mu_error_status, MU_EINVAL, 
		"Invalid filemode argument in mu_file_open_raw: %d",
		filemode);
    }

  /* Create the file structure */
  file = (struct mu_file *) mu_malloc (1,sizeof (struct mu_file));
  file->endian = -1;
  file->eof = 0;
  file->eof_handler = NULL;
  file->filemode = filemode;
  file->filename = (char *) mu_malloc (strlen (filename) + 1, sizeof (char));
  strcpy (file->filename, filename);

  /* Open the file */
  file->descriptor = open (filename, flags, 0600);
  if (file->descriptor < 0)
    {
      mu_error (mu_error_status, MU_EIO, 
		"Opening %s for %s failed: %s", 
		filename, 
		((filemode == mu_file_for_writing) ? "writing" :
		 ((filemode == mu_file_for_reading) ? "reading" :
		  ((filemode == mu_file_for_appending) ? "appending" :
		   "invalid file mode (oops!)"))),
		mu_libc_strerror(errno));
    }
  return file;
}

void mu_file_set_endian (struct mu_file *file, int endian)
{
  mu_endian_validate (endian, "mu_file_set_endian");
  file->endian = endian;
  file->swap_endian = ((endian != mu_endian_detect ()) ? 1 : 0);
}

void mu_file_assert_endian (struct mu_file *file, char *routine)
{
  if (file->endian == -1)
    {
      mu_error (mu_error_status, MU_EIO, 
		"No endian specified for open file %s", file->filename);
    }
}

static int mu_file_open_detect_endian (struct mu_file *file)
{
  int endian;
  if (mu_file_read_cookie (file, "endian-cookie"))
    {
      unsigned char endian_char;
      mu_file_read (file, &endian_char, 1);
      if (endian_char == 'l') 
	{ 
	  return mu_file_little_endian;
	}
      else if (endian_char == 'b') 
	{ 
	  return mu_file_big_endian;
	}
      else 
	{
	  mu_error (mu_error_status, MU_EIO,
		    "File %s has a corrupt endian cookie", file->filename);
	}
    }
  else
    {
      mu_error (mu_error_status, MU_EIO,
		"File %s does not have an endian cookie", file->filename);
    }
}

static void mu_file_open_write_endian_cookie (struct mu_file *file)
{
  unsigned char little_endian_char  = 'l';
  unsigned char big_endian_char     = 'b';

  mu_file_write_cookie (file, "endian-cookie");
  if (mu_endian_detect () == mu_file_big_endian)
    {
      mu_file_write (file, &big_endian_char, 1);
    }
  else
    {
      mu_file_write (file, &little_endian_char, 1);
    }
}


/* Open the file and handle endian-ness automatically by writing a
   cookie to the file */
struct mu_file *mu_file_open (char *filename, int filemode)
{
  struct mu_file *file;

  /* If we are opening the file for reading, then read it and 
     process the endianness cookie */
  if (filemode == mu_file_for_reading)
    {
      file = mu_file_open_raw (filename, mu_file_for_reading);
      mu_file_set_endian (file, mu_file_open_detect_endian (file));
    }
  /* If we are writing/creating a file, then open and record the cookie */
  else if (filemode == mu_file_for_writing)
    {
      file = mu_file_open_raw (filename, mu_file_for_writing);
      mu_file_set_endian (file, mu_endian_detect ());
      mu_file_open_write_endian_cookie (file);
    }
  /* If we are appending to a file, then open for reading first, detect
     the cookie, close and then open again for writing */
  else if (filemode == mu_file_for_appending)
    {
      /* Detect the endian cookie of this file */
      int endian;
      file = mu_file_open_raw (filename, mu_file_for_reading);
      endian = mu_file_open_detect_endian (file);
      mu_file_close (file);
      
      /* Open the file again for appending */
      file = mu_file_open_raw (filename, mu_file_for_appending);
      mu_file_set_endian (file, endian);
    }
  else
    {
      mu_error (mu_error_status, MU_EINVAL, 
		"Invalid filemode while opening %s", filename);
    }
  
  return file;
}

/* Close the file descriptor. This must be done very carefully */
void mu_file_close (struct mu_file *file)
{
  /* Close the file descriptor */
 try_again:
  if (close (file->descriptor) < 0)
    {
#ifdef EINTR
      if (errno == EINTR) goto try_again;
#endif
      mu_error (mu_error_status, MU_EIO, "Can't close file %s: %s", 
		file->filename, mu_libc_strerror (errno));
    }
  else
    {
      mu_free (file->filename);
      mu_free (file);
      return;
    }
}

/*
 * ----------------------------
 * Low-level input and output
 * ----------------------------
 */

/* The following routines have been adapted from fileutils which
   is licensed under the GNU GPL */

/* This is the robust way for writing and reading data from a file
   descriptor. The EINTR error is recoverable and is likely to happen
   in NFS mounted file systems. It is also recoverable if the read/write
   calls return with an indication that they have not written or received
   the whole data. Other errors are not recoverable. */

void mu_file_read (struct mu_file *file, void *p, size_t len)
{
  int total_read = 0;
  char *ptr = (char *) p;

  while (len > 0)
    {
      int n_chars = read (file->descriptor, ptr, len);
      /* If end-of-file */
      if (n_chars == 0)
	{
	  /* Set the end-of-file flag */
	  file->eof = 1;
	  /* Invoke eof callback, if one has been installed */
	  if (file->eof_handler) (*(file->eof_handler)) (file);
	  return;
	}
      if (n_chars < 0)
	{
#ifdef EINTR
	  if (errno == EINTR) continue;
#endif
	  mu_error (mu_error_status, MU_EIO,
		    "Failed reading data to file %s: %s",
		    file->filename, mu_libc_strerror (errno));
	}
      total_read += n_chars;
      ptr += n_chars;
      len -= n_chars;
    }
  return;
}

void mu_file_write (struct mu_file *file, void *p, size_t len)
{
  int total_written = 0;
  char *ptr = (char *) p;

  while (len > 0)
    {
      int written = write (file->descriptor, ptr, len);
      if (written < 0)
	{
#ifdef EINTR
	  if (errno == EINTR) continue;
#endif
	  mu_error (mu_error_status, MU_EIO,
		    "Failed writing data to file %s: %s",
		    file->filename, mu_libc_strerror (errno));
	}
      total_written += written;
      ptr += written;
      len -= written;
    }
  return;
}

/*
 * -----------------------------------
 * Dealing with end-of-file conditions
 * -----------------------------------
 */

/* Return 1 if end-of-file has been triggered */
int mu_file_eof_p (struct mu_file *file)
{
  return file->eof;
}

/* Set the eof handler. Return the old handler */
mu_file_eof_handler_t *
mu_file_set_eof_handler (struct mu_file *file, mu_file_eof_handler_t *handler)
{
  mu_file_eof_handler_t *old_handler = file->eof_handler;
  file->eof_handler = handler;
  return old_handler;
}

/* Set the handler to default */
static void mu_file_the_default_eof_handler (struct mu_file *file)
{
  mu_error (mu_error_status, MU_EIO, 
	    "Unexpected end-of-file while reading %s", file->filename);
}


mu_file_eof_handler_t *
mu_file_set_default_eof_handler (struct mu_file *file)
{
  return mu_file_set_eof_handler (file, mu_file_the_default_eof_handler);
}

/*
 * -------------
 * File cookies
 * -------------
 */
 
void mu_file_write_cookie (struct mu_file *file, char *cookie)
{
  /* Include terminating null on output */
  mu_file_write (file, cookie, strlen(cookie)+1);
}

int mu_file_read_cookie (struct mu_file *file, char *cookie)
{
  /* Buffer for reading in cookies. This buffer persists between calls */
  static unsigned char *buffer = NULL;
  static int buffer_length = 0;

  /* Get the length of the incoming cookie and allocate
     sufficient space for loading it in */
  int len = strlen (cookie) + 1;
  if (len > buffer_length)
    {
      buffer = mu_realloc (buffer, len, sizeof (unsigned char));
      buffer_length = len;
    }
  
  /* Read in the cookie */
  mu_file_read (file, buffer, len);

  /* Check if the cookie is null terminated */
  if (buffer[len-1] != '\0') return 0;

  /* Check if the cookie is equal to the given cookie */
  if (strcmp (cookie, buffer) != 0) return 0;

  /* Otherwise, everything is okey */
  return 1;
}

/*
 * -------------------------------
 * Medium-level input and output
 * -------------------------------
 */

/* Output routines */
void mu_file_write_double (struct mu_file *file, double *a, size_t n)
{
  mu_file_assert_endian (file, "mu_file_write_double");
  if (file->swap_endian) mu_endian_swap8_block (a, n);
  mu_file_write (file, a, n*sizeof (double));
  if (file->swap_endian) mu_endian_swap8_block (a, n);
}

void mu_file_write_float (struct mu_file *file, float *a, size_t n)
{
  mu_file_assert_endian (file, "mu_file_write_float");
  if (file->swap_endian) mu_endian_swap4_block (a, n);
  mu_file_write (file, a, n*sizeof (float));
  if (file->swap_endian) mu_endian_swap4_block (a, n);
}

void mu_file_write_int (struct mu_file *file, int *a, size_t n)
{
  mu_file_assert_endian (file, "mu_file_write_int");
  if (file->swap_endian) mu_endian_swap4_block (a, n);
  mu_file_write (file, a, n*sizeof (int));
  if (file->swap_endian) mu_endian_swap4_block (a, n);
}

/* Input routines */
void mu_file_read_double (struct mu_file *file, double *a, size_t n)
{
  mu_file_assert_endian (file, "mu_file_read_double");
  mu_file_read (file, a, n*sizeof (double));
  if (file->swap_endian) mu_endian_swap8_block (a, n);
}
void mu_file_read_float (struct mu_file *file, float *a, size_t n)
{
  mu_file_assert_endian (file, "mu_file_read_float");
  mu_file_read (file, a, n*sizeof (float));
  if (file->swap_endian) mu_endian_swap4_block (a, n);
}
void mu_file_read_int (struct mu_file *file, int *a, size_t n)
{
  mu_file_assert_endian (file, "mu_file_read_int");
  mu_file_read (file, a, n*sizeof (int));
  if (file->swap_endian) mu_endian_swap4_block (a, n);
}

/*
 * -------------------------
 * High level I/O functions
 * -------------------------
 */

/* The default input and output handlers */
static int mu_file_io_default_input_handler (void *clientdata, char *spec, 
					     va_list *arg)
{
  struct mu_file *file = (struct mu_file *) clientdata;

  /* Handlers for simple types */
  if (strcmp (spec, "i") == 0)
    {
      mu_file_read_int (file, va_arg (*arg, int *), 1);
      return 1;
    }
  if (strcmp (spec, "f") == 0)
    {
      mu_file_read_float (file, va_arg (*arg, float *), 1);
      return 1;
    }
  if (strcmp (spec, "d") == 0)
    {
      mu_file_read_double (file, va_arg (*arg, double *), 1);
      return 1;
    }

  /* Handlers for arrays */
  if ((strcmp (spec, "ia") == 0) || (strcmp (spec, "iaL") == 0))
    {
      mu_int_array_read (file, va_arg (*arg, struct mu_int_array *), 
			 ((spec[2] == 'L') ? 1 : 0));
      return 1;
    }
  if ((strcmp (spec, "fa") == 0) || (strcmp (spec, "faL") == 0))
    {
      mu_float_array_read (file, va_arg (*arg, struct mu_float_array *), 
			   ((spec[2] == 'L') ? 1 : 0));
      return 1;
    }
  if ((strcmp (spec, "da") == 0) || (strcmp (spec, "daL") == 0))
    {
      mu_double_array_read (file, va_arg (*arg, struct mu_double_array *),
			    ((spec[2] == 'L') ? 1 : 0));
      return 1;
    }

  /* Handlers for matrices */
  if ((strcmp (spec, "im") == 0) || (strcmp (spec, "imL") == 0))
    {
      mu_int_matrix_read (file, va_arg (*arg, struct mu_int_matrix *),
			  ((spec[2] == 'L') ? 1 : 0));
      return 1;
    }
  if ((strcmp (spec, "fm") == 0) || (strcmp (spec, "fmL") == 0))
    {
      mu_float_matrix_read (file, va_arg (*arg, struct mu_float_matrix *),
			    ((spec[2] == 'L') ? 1 : 0));
      return 1;
    }
  if ((strcmp (spec, "dm") == 0) || (strcmp (spec, "dmL") == 0))
    {
      mu_double_matrix_read (file, va_arg (*arg, struct mu_double_matrix *),
			     ((spec[2] == 'L') ? 1 : 0));
      return 1;
    }

  /* Handlers for cubes */
  if ((strcmp (spec, "ic") == 0) || (strcmp (spec, "icL") == 0))
    {
      mu_int_cube_read (file, va_arg (*arg, struct mu_int_cube *),
			((spec[2] == 'L') ? 1 : 0));
      return 1;
    }
  if ((strcmp (spec, "fc") == 0) || (strcmp (spec, "fcL") == 0))
    {
      mu_float_cube_read (file, va_arg (*arg, struct mu_float_cube *),
			  ((spec[2] == 'L') ? 1 : 0));
      return 1;
    }
  if ((strcmp (spec, "dc") == 0) || (strcmp (spec, "dcL") == 0))
    {
      mu_double_cube_read (file, va_arg (*arg, struct mu_double_cube *),
			   ((spec[2] == 'L') ? 1 : 0));
      return 1;
    }

  return 0;
}
static int mu_file_io_default_output_handler (void *clientdata, char *spec,
					      va_list *arg)
{
  struct mu_file *file = (struct mu_file *) clientdata;

  /* Handlers for simple types */
  if (strcmp (spec, "i") == 0)
    {
      int value = va_arg (*arg, int);
      mu_file_write_int (file, &value, 1);
      return 1;
    }
  if (strcmp (spec, "f") == 0)
    {
      float value = (float) va_arg (*arg, double); /* handle type promotion */
      mu_file_write_float (file, &value, 1); 
      return 1;
    }
  if (strcmp (spec, "d") == 0)
    {
      double value = va_arg (*arg, double);
      mu_file_write_double (file, &value, 1);
      return 1;
    }

  /* Handlers for arrays */
  if (strcmp (spec, "ia") == 0)
    {
      mu_int_array_write (file, va_arg (*arg, struct mu_int_array));
      return 1;
    }
  if (strcmp (spec, "fa") == 0)
    {
      mu_float_array_write (file, va_arg (*arg, struct mu_float_array));
      return 1;
    }
  if (strcmp (spec, "da") == 0)
    {
      mu_double_array_write (file, va_arg (*arg, struct mu_double_array));
      return 1;
    }
  
  /* Handlers for matrices */
  if (strcmp (spec, "im") == 0)
    {
      mu_int_matrix_write (file, va_arg (*arg, struct mu_int_matrix));
      return 1;
    }
  if (strcmp (spec, "fm") == 0)
    {
      mu_float_matrix_write (file, va_arg (*arg, struct mu_float_matrix));
      return 1;
    }
  if (strcmp (spec, "dm") == 0)
    {
      mu_double_matrix_write (file, va_arg (*arg, struct mu_double_matrix));
      return 1;
    }
  
  /* Handlers for cubes */
  if (strcmp (spec, "ic") == 0)
    {
      mu_int_cube_write (file, va_arg (*arg, struct mu_int_cube));
      return 1;
    }
  if (strcmp (spec, "fc") == 0)
    {
      mu_float_cube_write (file, va_arg (*arg, struct mu_float_cube));
      return 1;
    }
  if (strcmp (spec, "dc") == 0)
    {
      mu_double_cube_write (file, va_arg (*arg, struct mu_double_cube));
      return 1;
    }
  
  return 0;
}

/* Tables containing the current handlers */
static struct mu_var_handler_table mu_file_io_input_handlers[] =
{
  { mu_file_io_default_input_handler, NULL }
};

static struct mu_var_handler_table mu_file_io_output_handlers[] = 
{
  { mu_file_io_default_output_handler, NULL }
};

/* Register new handler pairs */
void mu_file_register (mu_var_handler_t *input, mu_var_handler_t *output)
{
  mu_var_register (mu_file_io_input_handlers, input);
  mu_var_register (mu_file_io_output_handlers, output);
}

/* Output */
void mu_file_output (struct mu_file *file, char *fmt, ...)
{
  va_list args;
  VA_START (args, fmt);
  mu_var_call (mu_file_io_output_handlers, file, 
	       "mu_file_output", fmt, &args);
  va_end (args);
}

/* Input */
void mu_file_input (struct mu_file *file, char *fmt, ...)
{
  va_list args;
  VA_START (args, fmt);
  mu_var_call (mu_file_io_input_handlers, file, 
	       "mu_file_input", fmt, &args);
  va_end (args);
}
