/*
** Copyright (C) 1999 Eleftherios Gkioulekas <lf@amath.washington.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
# include <stdlib.h>
#endif
#if HAVE_MALLOC_H
# include <malloc.h>
#endif

#include <mu.h>

/* The null memory buffer */
struct mu_membuf mu_membuf_null = { NULL, 0, NULL };

/* Memory buffers should be created with one of the following two methods
   before they are ever used */

/* Initialize the memory buffer. This does not allocate memory but it
   allows you to use the reallocate and reattach functions on it */
void mu_membuf_initialize (struct mu_membuf *buf)
{
  buf->data = NULL;
  buf->size = 0;
  buf->refcount = NULL;
}

/* Return a memory buffer that is bound to a newly allocated region */
struct mu_membuf mu_membuf_allocate (size_t size)
{
  struct mu_membuf buf;

  /* Allocate memory region */
  buf.data = mu_calloc (size, 1);
  buf.size = size;

  /* Initialize the reference counter */
  buf.refcount = (size_t *) mu_malloc (1, sizeof(size_t));
  *(buf.refcount) = 1;
  return buf;
}

/* Return a memory buffer that is bound to the same region as another 
   buffer */
struct mu_membuf mu_membuf_attach_to (struct mu_membuf parent)
{
  struct mu_membuf buf;

  /* Get the same memory size and reference counting as the other buffer */
  buf.data = parent.data;
  buf.size = parent.size;
  buf.refcount = parent.refcount;

  /* Increase the reference count */
  *(buf.refcount) += 1;
  return buf;
}


/* Release the memory buffer from it's current binding. If all memory
   buffers bound to the same region release, then the region is dealloced */
void mu_membuf_release (struct mu_membuf buf)
{
  /* If we are attached to something */
  if (buf.refcount != NULL && buf.data != NULL)
    {
      /* Decrease the reference counter */
      *(buf.refcount) -= 1;
      /* If the reference couter is zero, then collect the garbage */
      if (*(buf.refcount) == 0)
	{
	  mu_free (buf.refcount);
	  mu_free (buf.data);
	}
      buf.data = NULL;
      buf.refcount = NULL;
      buf.size = 0;
    }
}

/* Rebind an already bound buffer to a new region */
void mu_membuf_reallocate (struct mu_membuf *buf, size_t size)
{
  if (buf->refcount != NULL) mu_membuf_release (*buf);
  *buf = mu_membuf_allocate (size);
}

/* Rebind an already bound buffer to another buffer */
void mu_membuf_reattach_to (struct mu_membuf *buf, struct mu_membuf parent)
{
  if (buf->refcount != NULL) mu_membuf_release (*buf);
  *buf = mu_membuf_attach_to (parent);
}

/* Return 0 if the memory buffer is not bound. Return the number of
   references if the buffer is bound */
int mu_membuf_references (struct mu_membuf buf)
{
  return ((buf.refcount != NULL) ? (*(buf.refcount)) : 0);
}

/* Return the address bound to this memory buffer */
void *mu_membuf_address (struct mu_membuf buf)
{
  return buf.data;
}
