/*
** Copyright (C) 1999 Eleftherios Gkioulekas <lf@amath.washington.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
# include <stdlib.h>
#endif
#if HAVE_MALLOC_H
# include <malloc.h>
#endif
#include <mu.h>
#include <assert.h>

typedef struct mu_double_array double_array;
typedef struct mu_double_matrix double_matrix;
typedef struct mu_double_cube double_cube;

/* Test the arrays */
void test_array (void)
{
  double_array a,b;
  size_t N = 1000;
  size_t i;

  /* Set up array a and check the setup */
  a = mu_double_array_alloc (N);
  assert (mu_double_array_contiguous_p (a));
  loop_array (i,a) A_(a,i) = ((double) i);
  loop (i,0,a.X-1) assert (A_(a,i) == ((double) i));

  /* Check whether aliases work */
  b = mu_double_array_alias (a);
  loop_array (i,b) assert (A_(b,i) == ((double) i));
  loop_array (i,b) A_(b,i) = 12;
  loop_array (i,a) assert (A_(a,i) == 12);

  /* The memory buffer used by a and b should have 2 references */
  assert (mu_membuf_references (a.buf) == 2);

  /* Reallocate the array a */
  mu_double_array_realloc (&a,2*N);
  loop_array (i,a) A_(a,i) = ((double) i)+1;
  loop_array (i,b) assert (A_(b,i) == 12);
  mu_double_array_realias (&b, a);
  loop_array (i,b) assert (A_(b,i) == ((double) i)+1);

  /* If we free the alias, there should be only 1 reference */
  mu_double_array_free (a);
  assert (mu_membuf_references (b.buf) == 1);
  mu_double_array_free (b);
}

/* Test the matrix */
void test_matrix (void)
{
  double_matrix a,b;
  size_t i,j;
  size_t N1 = 200;
  size_t N2 = 100;

  /* Set up matrix a and check the setup */
  a = mu_double_matrix_alloc (N1,N2);
  assert (mu_double_matrix_contiguous_p (a));
  loop_matrix (i,j,a) M_(a,i,j) = ((double) (i+j));
  loop (i,0,a.X-1) loop(j,0,a.Y-1) assert (M_(a,i,j) == ((double) (i+j)));
  
  /* Check whether aliases work */
  b = mu_double_matrix_alias (a);
  loop_matrix (i,j,b) assert (M_(b,i,j) == ((double) (i+j)));
  loop_matrix (i,j,b) M_(b,i,j) = 12;
  loop_matrix (i,j,a) assert (M_(a,i,j) == 12);

  /* Reallocate the matrix a */
  mu_double_matrix_realloc (&a, 2*N1, 2*N2);
  loop_matrix (i,j,a) M_(a,i,j) = ((double) i)+1;
  loop_matrix (i,j,b) assert (M_(b,i,j) == 12);
  mu_double_matrix_realias (&b, a);
  loop_matrix (i,j,b) assert (M_(b,i,j) == ((double) i) +1);

  /* Free the matrices */
  mu_double_matrix_free (a);
  assert (mu_membuf_references (b.buf) == 1);
  mu_double_matrix_free (b);
}

/* Test the cube */
void test_cube (void)
{
  double_cube a,b;
  size_t i,j,k;
  size_t N1 = 20;
  size_t N2 = 24;
  size_t N3 = 22;

  /* Set up a cube a and check the setup */
  a = mu_double_cube_alloc (N1,N2,N3);
  assert (mu_double_cube_contiguous_p (a));
  loop_cube (i,j,k,a) C_(a,i,j,k) = ((double) (i+j+k));
  loop (i,0,a.X-1) loop (j,0,a.Y-1) loop (k,0,a.Z-1) 
    assert (C_(a,i,j,k) == ((double) (i+j+k)));

  /* Check whether aliases work */
  b = mu_double_cube_alias (a);
  loop_cube (i,j,k,b) assert (C_(b,i,j,k) == ((double) (i+j+k)));
  loop_cube (i,j,k,b) C_(b,i,j,k) = 12;
  loop_cube (i,j,k,a) assert (C_(a,i,j,k) == 12);

  /* Reallocate the cube a */
  mu_double_cube_realloc (&a, 2*N1, 2*N2, 2*N3);
  loop_cube (i,j,k,a) C_(a,i,j,k) = ((double) i)+1;
  loop_cube (i,j,k,b) assert (C_(b,i,j,k) == 12);
  mu_double_cube_realias (&b, a);
  loop_cube (i,j,k,b) assert (C_(b,i,j,k) == ((double) i)+1);
  
  /* Free the cubes */
  mu_double_cube_free (a);
  assert (mu_membuf_references (b.buf) == 1);
  mu_double_cube_free (b);
}

/* Test subarrays */
void test_sub_array (void)
{
  double_array a;
  double_array b = mu_double_array_null;
  size_t N = 1000;
  size_t i1 = 10;
  size_t i2 = 100;
  size_t i;
  
  /* Set up array and subarray */
  a = mu_double_array_alloc (N);
  mu_double_array_sub (&b, a, i1, i2);

  /* Assign values to a */
  loop_array (i,a) A_(a,i) = ((double) i);
  loop_array (i,b) assert (A_(b,i) == A_(a,(i+i1)));
  
  /* Free the arrays */
  mu_double_array_free (a);
  mu_double_array_free (b);
}

/* Test submatrices */
void test_sub_matrix (void)
{
  double_matrix a;
  double_matrix b = mu_double_matrix_null;
  size_t N1 = 60; size_t N2 = 40; 
  size_t i1 = 10; size_t i2 = 30; size_t j1 = 20; size_t j2 = 25;
  size_t i,j;
  
  /* Setup matrix and submatrix */
  a = mu_double_matrix_alloc (N1,N2);
  mu_double_matrix_sub (&b, a, i1, i2, j1, j2);
  
  /* Test the submatrix */
  loop_matrix (i,j,a) M_(a,i,j) = ((double) i + 2*j);
  loop_matrix (i,j,b) assert (M_(b,i,j) == M_(a,(i+i1),(j+j1)));
  
  /* Free the matrices */
  mu_double_matrix_free (a);
  mu_double_matrix_free (b);
}

/* Test subcubes */
void test_sub_cube (void)
{
  double_cube a;
  double_cube b = mu_double_cube_null;
  size_t N1, N2, N3, i1, i2, j1, j2, k1, k2;
  size_t i,j,k;

  /* Setup the sizes */
  N1 = 20; N2 = 22; N3 = 25;
  i1 = 5; i2 = 15;
  j1 = 4; j2 = 18;
  k1 = 6; k2 = 20;

  /* Initialize the cubes */
  a = mu_double_cube_alloc (N1,N2,N3);
  b = mu_double_cube_initialize ();
  mu_double_cube_sub (&b, a, i1, i2, j1, j2, k1, k2);

  /* Test out if it works */
  loop_cube (i,j,k,a) C_(a,i,j,k) = ((double) i+2*j+3*k);
  loop_cube (i,j,k,b) assert (C_(b,i,j,k) == C_(a,(i+i1),(j+j1),(k+k1)));

  /* Free the cubes */
  mu_double_cube_free (a);
  mu_double_cube_free (b);
}

void test_interlaced_array (void)
{
  double_array a, a1, a2;
  size_t N = 100;
  size_t i;

  /* Initialize the arrays */
  a = mu_double_array_alloc (N);
  a1 = a2 = mu_double_array_null;
  mu_double_array_interlaced (a, &a1, &a2);

  /* Test out if it works */
  loop_array (i,a1) A_(a1,i) = ((double) i);
  loop_array (i,a2) A_(a2,i) = ((double) 2*i);
  loop_array (i,a) assert (A_(a,i) == ((i%2 == 0) ? (i/2) : (i-1)));
  
  /* Free the arrays */
  mu_double_array_free (a);
  mu_double_array_free (a1);
  mu_double_array_free (a2);
}

void test_interlaced_matrix (void)
{
  double_matrix a, a1, a2;
  size_t N1,N2,i,j;

  /* Initialize the matrices */
  N1 = 50; N2 = 30;
  a = mu_double_matrix_alloc (N1, N2);
  a1 = a2 = mu_double_matrix_null;
  mu_double_matrix_interlaced (a, &a1, &a2);

  /* Test out if it works */
  loop_matrix (i,j,a1) M_(a1,i,j) = ((double) i+j);
  loop_matrix (i,j,a2) M_(a2,i,j) = ((double) 2*i+j);
  loop_matrix (i,j,a) assert (M_(a,i,j) == (j+((i%2 == 0) ? (i/2) : (i-1))));

  /* Free the matrices */
  mu_double_matrix_free (a);
  mu_double_matrix_free (a1);
  mu_double_matrix_free (a2);
}

void test_interlaced_cube (void)
{
  double_cube a, a1, a2;
  size_t N1,N2,N3,i,j,k;
  
  /* Initielize the matrices */
  N1 = 30; N2 = 20; N3 = 25;
  a = mu_double_cube_alloc (N1, N2, N3);
  a1 = a2 = mu_double_cube_null;
  mu_double_cube_interlaced (a, &a1, &a2);
  
  /* Test out if it works */
  loop_cube (i,j,k,a1) C_(a1,i,j,k) = ((double) i+j+3*k);
  loop_cube (i,j,k,a2) C_(a2,i,j,k) = ((double) 2*i+j+3*k);
  loop_cube (i,j,k,a) 
    assert (C_(a,i,j,k) == (j+3*k) + ((i%2 == 0) ? (i/2) : (i-1)));

  /* Free the cubes */
  mu_double_cube_free (a);
  mu_double_cube_free (a1);
  mu_double_cube_free (a2);
}

void test_matrix_to_array (void)
{
  const size_t N1 = 45;
  const size_t N2 = 30;
  size_t n = 0;
  size_t i,j;
  double_matrix A;
  double_array a;

  /* Initialize matrix and array slice */
  A = mu_double_matrix_alloc (N1, N2);
  loop_matrix (i,j,A) M_(A,i,j) = ((double) (n++));
  a = mu_double_array_null;
  
  /* Check X_slice */
  loop (j,0,N2-1)
    {
      mu_double_matrix_to_array (&a, A, X_slice, j);
      loop (i,0,N1-1) assert (A_(a,i) == M_(A,i,j));
    }
  
  /* Check Y_slice */
  loop (i,0,N1-1)
    {
      mu_double_matrix_to_array (&a, A, Y_slice, i);
      loop (j,0,N2-1) assert (A_(a,j) == M_(A,i,j));
    }
  
  /* Free the arrays */
  mu_double_matrix_free (A);
  mu_double_array_free (a);
}

void test_cube_to_array (void)
{
  const size_t N1 = 30;
  const size_t N2 = 40;
  const size_t N3 = 25;
  double n = 0;
  size_t i,j,k;
  double_cube A;
  double_array a;

  /* Initialize cube and array slice */
  A = mu_double_cube_alloc (N1, N2, N3);
  loop_cube (i,j,k,A) C_(A,i,j,k) = n++;
  a = mu_double_array_null;

  /* Check the X_slice */
  loop (j,0,N2-1) loop (k,0,N3-1)
    {
      mu_double_cube_to_array (&a, A, X_slice, j, k);
      loop (i,0,N1-1) assert (A_(a,i) == C_(A,i,j,k));
    }

  /* Check the Y_slice */
  loop (i,0,N1-1) loop (k,0,N3-1)
    {
      mu_double_cube_to_array (&a, A, Y_slice, i, k);
      loop (j,0,N2-1) assert (A_(a,j) == C_(A,i,j,k));
    }

  /* Check the Z_slice */
  loop (i,0,N1-1) loop (j,0,N2-1)
    {
      mu_double_cube_to_array (&a, A, Z_slice, i, j);
      loop (k,0,N3-1) assert (A_(a,k) == C_(A,i,j,k));
    }

  /* Free the arrays */
  mu_double_cube_free (A);
  mu_double_array_free (a);
}

void test_cube_to_matrix (void)
{
  const size_t N1 = 30;
  const size_t N2 = 40;
  const size_t N3 = 25;
  double n = 0;
  size_t i,j,k;
  double_cube A;
  double_matrix a;
  
  /* Initialize cube and matrix slice */
  A = mu_double_cube_alloc (N1, N2, N3);
  loop_cube (i,j,k,A) C_(A,i,j,k) = n++;
  a = mu_double_matrix_null;
  
  /* Check the XY_slice */
  loop (k,0,N3-1)
    {
      mu_double_cube_to_matrix (&a, A, XY_slice, k);
      loop_matrix (i,j,a) assert (M_(a,i,j) == C_(A,i,j,k));
    }

  /* Check the XZ_slice */
  loop (j,0,N2-1)
    {
      mu_double_cube_to_matrix (&a, A, XZ_slice, j);
      loop_matrix (i,k,a) assert (M_(a,i,k) == C_(A,i,j,k));
    }
  
  /* Check the YX_slice */
  loop (k,0,N3-1)
    {
      mu_double_cube_to_matrix (&a, A, YX_slice, k);
      loop_matrix (j,i,a) assert (M_(a,j,i) == C_(A,i,j,k));
    }

  /* Check the YZ_slice */
  loop (i,0,N1-1)
    {
      mu_double_cube_to_matrix (&a, A, YZ_slice, i);
      loop_matrix (j,k,a) assert (M_(a,j,k) == C_(A,i,j,k));
    }
  
  /* Check the ZX_slice */
  loop (j,0,N2-1)
    {
      mu_double_cube_to_matrix (&a, A, ZX_slice, j);
      loop_matrix (k,i,a) assert (M_(a,k,i) == C_(A,i,j,k));
    }
  
  /* Check the ZY_slice */
  loop (i,0,N1-1)
    {
      mu_double_cube_to_matrix (&a, A, ZY_slice, i);
      loop_matrix (k,j,a) assert (M_(a,k,j) == C_(A,i,j,k));
    }

  /* Free the arrays */
  mu_double_cube_free (A);
  mu_double_matrix_free (a);
}

void test_array_free_many (void)
{
  const int N = 20;
  double_array a1, a2, a3, a4, a5;
  a1 = mu_double_array_alloc (N);
  a2 = mu_double_array_alloc (N);
  a3 = mu_double_array_alloc (N);
  a4 = mu_double_array_alloc (N);
  a5 = mu_double_array_alloc (N);
  mu_double_array_free_many (5, a1, a2, a3, a4, a5);
}

void test_matrix_free_many (void)
{
  const int N = 20;
  double_matrix a1, a2, a3, a4, a5;
  a1 = mu_double_matrix_alloc (N, N);
  a2 = mu_double_matrix_alloc (N, N);
  a3 = mu_double_matrix_alloc (N, N);
  a4 = mu_double_matrix_alloc (N, N);
  a5 = mu_double_matrix_alloc (N, N);
  mu_double_matrix_free_many (5, a1, a2, a3, a4, a5);
}

void test_cube_free_many (void)
{
  const int N = 20;
  double_cube a1, a2, a3, a4, a5;
  a1 = mu_double_cube_alloc (N, N, N);
  a2 = mu_double_cube_alloc (N, N, N);
  a3 = mu_double_cube_alloc (N, N, N);
  a4 = mu_double_cube_alloc (N, N, N);
  a5 = mu_double_cube_alloc (N, N, N);
  mu_double_cube_free_many (5, a1, a2, a3, a4, a5);
}

int main ()
{
  int i;

  loop (i,0,10) test_array ();
  loop (i,0,10) test_matrix ();
  loop (i,0,10) test_cube ();

  loop (i,0,10) test_sub_array ();
  loop (i,0,10) test_sub_matrix ();
  loop (i,0,10) test_sub_cube ();

  loop (i,0,10) test_interlaced_array ();
  loop (i,0,10) test_interlaced_matrix ();
  loop (i,0,10) test_interlaced_cube ();

  loop (i,0,10) test_matrix_to_array ();
  loop (i,0,10) test_cube_to_array ();
  loop (i,0,10) test_cube_to_matrix ();

  loop (i,0,10) test_array_free_many ();
  loop (i,0,10) test_matrix_free_many ();
  loop (i,0,10) test_cube_free_many ();
  exit (0);
}

